@@ -1,4 +1,4 @@
-/* Copyright (C) 1995,1996,1997,1998,1999,2000,2001, 2002, 2003, 2006 Free Software Foundation, Inc.
+/* Copyright (C) 1995,1996,1997,1998,1999,2000,2001, 2002, 2003, 2006, 2008 Free Software Foundation, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -15,11 +15,9 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
-#define _GNU_SOURCE
-
 /* #define DEBUGINFO */
 
-#if HAVE_CONFIG_H
+#ifdef HAVE_CONFIG_H
 #  include <config.h>
 #endif
 
@@ -28,11 +26,6 @@
 #include <string.h>
 #include <assert.h>
 
-#ifdef __ia64__
-#include <ucontext.h>
-extern unsigned long * __libc_ia64_register_backing_store_base;
-#endif
-
 #include "libguile/_scm.h"
 #include "libguile/eval.h"
 #include "libguile/stime.h"
@@ -213,6 +206,7 @@ unsigned long scm_mtrigger;
 /* GC Statistics Keeping
  */
 unsigned long scm_cells_allocated = 0;
+unsigned long scm_last_cells_allocated;
 unsigned long scm_mallocated = 0;
 unsigned long scm_gc_cells_collected;
 unsigned long scm_gc_cells_collected_1 = 0; /* previous GC yield */
@@ -225,6 +219,7 @@ unsigned long scm_gc_times = 0;
 unsigned long scm_gc_cells_swept = 0;
 double scm_gc_cells_marked_acc = 0.;
 double scm_gc_cells_swept_acc = 0.;
+double scm_gc_cells_allocated_acc = 0.;
 int scm_gc_cell_yield_percentage =0;
 int scm_gc_malloc_yield_percentage = 0;
 unsigned long protected_obj_count = 0;
@@ -243,6 +238,7 @@ SCM_SYMBOL (sym_cells_swept, "cells-swept");
 SCM_SYMBOL (sym_malloc_yield, "malloc-yield");
 SCM_SYMBOL (sym_cell_yield, "cell-yield");
 SCM_SYMBOL (sym_protected_objects, "protected-objects");
+SCM_SYMBOL (sym_total_cells_allocated, "total-cells-allocated");
 
 
 
@@ -315,6 +311,7 @@ SCM_DEFINE (scm_gc_stats, "gc-stats", 0, 0, 0,
   unsigned long int local_protected_obj_count;
   double local_scm_gc_cells_swept;
   double local_scm_gc_cells_marked;
+  double local_scm_total_cells_allocated;
   SCM answer;
   unsigned long *bounds = 0;
   int table_size = scm_i_heap_segment_table_size;  
@@ -356,6 +353,9 @@ SCM_DEFINE (scm_gc_stats, "gc-stats", 0, 0, 0,
     +(double) scm_gc_cells_swept 
     -(double) scm_gc_cells_collected;
 
+  local_scm_total_cells_allocated = scm_gc_cells_allocated_acc
+    + (double) (scm_cells_allocated - scm_last_cells_allocated);
+
   for (i = table_size; i--;)
     {
       heap_segs = scm_cons (scm_cons (scm_from_ulong (bounds[2*i]),
@@ -369,6 +369,8 @@ SCM_DEFINE (scm_gc_stats, "gc-stats", 0, 0, 0,
 			  scm_from_ulong (local_scm_gc_time_taken)),
 		scm_cons (sym_cells_allocated,
 			  scm_from_ulong (local_scm_cells_allocated)),
+		scm_cons (sym_total_cells_allocated,
+			  scm_from_ulong (local_scm_total_cells_allocated)),
 		scm_cons (sym_heap_size,
 			  scm_from_ulong (local_scm_heap_size)),
 		scm_cons (sym_mallocated,
@@ -967,7 +969,7 @@ gc_async_thunk (void)
  */
 static void *
 mark_gc_async (void * hook_data SCM_UNUSED,
-	       void *func_data SCM_UNUSED,
+	       void *fn_data SCM_UNUSED,
 	       void *data SCM_UNUSED)
 {
   /* If cell access debugging is enabled, the user may choose to perform
@@ -1021,6 +1023,44 @@ scm_init_gc ()
 #include "libguile/gc.x"
 }
 
+#ifdef __ia64__
+# ifdef __hpux
+#  include <sys/param.h>
+#  include <sys/pstat.h>
+void *
+scm_ia64_register_backing_store_base (void)
+{
+  struct pst_vm_status vm_status;
+  int i = 0;
+  while (pstat_getprocvm (&vm_status, sizeof (vm_status), 0, i++) == 1)
+    if (vm_status.pst_type == PS_RSESTACK)
+      return (void *) vm_status.pst_vaddr;
+  abort ();
+}
+void *
+scm_ia64_ar_bsp (const void *ctx)
+{
+  uint64_t bsp;
+  __uc_get_ar_bsp(ctx, &bsp);
+  return (void *) bsp;
+}
+# endif /* hpux */
+# ifdef linux
+#  include <ucontext.h>
+void *
+scm_ia64_register_backing_store_base (void)
+{
+  extern void *__libc_ia64_register_backing_store_base;
+  return __libc_ia64_register_backing_store_base;
+}
+void *
+scm_ia64_ar_bsp (const void *opaque)
+{
+  const ucontext_t *ctx = opaque;
+  return (void *) ctx->uc_mcontext.sc_ar_bsp;
+}
+# endif	/* linux */
+#endif /* __ia64__ */
 
 void
 scm_gc_sweep (void)
