@@ -1,4 +1,4 @@
-/* Copyright (C) 1995,1996,1998,1999,2000,2001, 2003, 2006 Free Software Foundation, Inc.
+/* Copyright (C) 1995,1996,1998,1999,2000,2001, 2003, 2006, 2008 Free Software Foundation, Inc.
  * 
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,6 +17,9 @@
 
 
 
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include <stdio.h>
 #include "libguile/_scm.h"
@@ -50,7 +53,7 @@ scm_c_hook_init (scm_t_c_hook *hook, void *hook_data, scm_t_c_hook_type type)
 void
 scm_c_hook_add (scm_t_c_hook *hook,
 		scm_t_c_hook_function func,
-		void *func_data, 
+		void *fn_data, 
 		int appendp)
 {
   scm_t_c_hook_entry *entry = scm_malloc (sizeof (scm_t_c_hook_entry));
@@ -60,19 +63,19 @@ scm_c_hook_add (scm_t_c_hook *hook,
       loc = &(*loc)->next;
   entry->next = *loc;
   entry->func = func;
-  entry->data = func_data;
+  entry->data = fn_data;
   *loc = entry;
 }
 
 void
 scm_c_hook_remove (scm_t_c_hook *hook,
 		   scm_t_c_hook_function func,
-		   void *func_data)
+		   void *fn_data)
 {
   scm_t_c_hook_entry **loc = &hook->first;
   while (*loc)
     {
-      if ((*loc)->func == func && (*loc)->data == func_data)
+      if ((*loc)->func == func && (*loc)->data == fn_data)
 	{
 	  scm_t_c_hook_entry *entry = *loc;
 	  *loc = (*loc)->next;
